/*-*- mode:c;indent-tabs-mode:nil;c-basic-offset:2;tab-width:8;coding:utf-8 -*-│
│ vi: set et ft=c ts=2 sts=2 sw=2 fenc=utf-8                               :vi │
╞══════════════════════════════════════════════════════════════════════════════╡
│ Copyright 2020 Justine Alexandra Roberts Tunney                              │
│                                                                              │
│ Permission to use, copy, modify, and/or distribute this software for         │
│ any purpose with or without fee is hereby granted, provided that the         │
│ above copyright notice and this permission notice appear in all copies.      │
│                                                                              │
│ THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL                │
│ WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED                │
│ WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE             │
│ AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL         │
│ DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR        │
│ PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER               │
│ TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR             │
│ PERFORMANCE OF THIS SOFTWARE.                                                │
╚─────────────────────────────────────────────────────────────────────────────*/
#include "libc/calls/struct/iovec.h"
#include "libc/macros.h"
#include "libc/sock/internal.h"
#include "libc/sysv/consts/iov.h"

/**
 * Converts I/O vector from System Five to WIN32 ABI.
 *
 * @return effective iovlen
 * @see IOV_MAX
 */
textwindows size_t __iovec2nt(struct NtIovec iovnt[hasatleast 16],
                              const struct iovec *iov, size_t iovlen) {
  size_t i, j, limit = 0x7ffff000;
  for (j = i = 0; i < iovlen; ++i) {
    if (!iov[i].iov_len)
      continue;
    if (j == 16)
      break;
    iovnt[j].buf = iov[i].iov_base;
    if (iov[i].iov_len < limit) {
      limit -= (iovnt[j].len = iov[i].iov_len);
      ++j;
    } else {
      iovnt[j].len = limit;
      ++j;
      break;
    }
  }
  return j;
}
